@@ -1,6 +1,9 @@
 # -*- coding: utf-8 -*-
 
-import argparse
+"""
+Secrets environment class and related variables, functions.
+"""
+
 import base64
 import binascii
 import hashlib
@@ -16,23 +19,31 @@ import re
 # directly here (rather than using 'from ... import') to avoid the conflict.
 import secrets
 #
-import stat
-import sys
-
 import uuid
 
 from collections import OrderedDict
+from pathlib import Path
 from shutil import copy
-from shutil import copytree
-from shutil import Error
-
+from stat import S_IMODE
 from xkcdpass import xkcd_password as xp
 
+from psec.exceptions import (
+    BasedirNotFoundError,
+    SecretNotFoundError,
+)
 from psec.utils import (
+    copydescriptions,
     find,
+    get_default_environment,
+    get_default_secrets_basedir,
     get_files_from_path,
-    get_fs_type,
+    is_secrets_basedir,
     remove_other_perms,
+    secrets_basedir_create,
+    DEFAULT_MODE,
+    Memoize,
+    SECRETS_DESCRIPTIONS_DIR,
+    SECRETS_FILE,
 )
 
 
@@ -52,6 +63,7 @@ BOOLEAN_OPTIONS = [
 ]
 
 DEFAULT_SIZE = 18
+# TODO(dittrich): Replace this with pydantic classes at some point.
 SECRET_TYPES = [
         OrderedDict({
             'Type': 'password',
@@ -105,12 +117,13 @@ SECRET_TYPES = [
 SECRET_ATTRIBUTES = [
     'Variable',
     'Group',
+    'Help',
     'Type',
     'Export',
     'Prompt',
     'Options'
 ]
-DEFAULT_MODE = 0o710
+
 # XKCD password defaults
 # See: https://www.unix-ninja.com/p/your_xkcd_passwords_are_pwned
 WORDS = 4
@@ -121,196 +134,6 @@ MAX_ACROSTIC_LENGTH = 6
 DELIMITER = '.'
 
 
-def natural_number(value):
-    """
-    Tests for a natural number.
-
-    Args:
-      value: The value to test
-
-    Returns:
-      A boolean indicating whether the value is a natural number or not.
-    """
-
-    ivalue = int(value)
-    if ivalue <= 0:
-        raise argparse.ArgumentTypeError(
-            f"[-] '{value}' is not a positive integer")
-    return ivalue
-
-
-# TODO(dittrich): Improve this?
-def _is_default(a, b):
-    """
-    Return "Yes" or "No" depending on whether e is the default
-    environment or not.
-    """
-    return "Yes" if a == b else "No"
-
-
-def get_local_default_file(cwd=None):
-    """Returns the path to the local identifier file."""
-    # TODO(dittrich): May need to do this differently to support
-    # Windows file systems.
-    if cwd is None:
-        cwd = os.getcwd()
-    return os.path.join(cwd, '.python_secrets_environment')
-
-
-def save_default_environment(
-    environment=None,
-    cwd=None
-):
-    """Save environment identifier to local file for defaulting."""
-    env_file = get_local_default_file(cwd=cwd)
-    with open(env_file, 'w') as f_out:
-        f_out.write('{0}\n'.format(str(environment)))
-    return True
-
-
-def clear_saved_default_environment(cwd=None):
-    """Remove saved default environment file."""
-    env_file = get_local_default_file(cwd=cwd)
-    if os.path.exists(env_file):
-        os.remove(env_file)
-        return True
-    else:
-        return False
-
-
-def get_saved_default_environment(cwd=None):
-    """Return environment ID value saved in local file or None."""
-    env_file = get_local_default_file(cwd=cwd)
-    saved_default = None
-    if os.path.exists(env_file):
-        with open(env_file, 'r') as f:
-            saved_default = f.read().replace('\n', '')
-    return saved_default
-
-
-def get_default_environment(cwd=None):
-    """
-    Return the default environment identifier.
-
-    There are multiple ways for a user to specify the environment
-    to use for python_secrets commands. Some of these involve
-    explicit settings (e.g., via command line option, a
-    saved value in the current working directory, or an
-    environment variable) or implicitly from the name of the
-    current working directory.
-    """
-
-    #  NOTE(dittrich): I know this code has multiple return points
-    #  but it is simpler and easier to understand this way.
-    #
-    # Highest priority is inhereted environment variable.
-    environment = os.getenv('D2_ENVIRONMENT', None)
-    if environment is not None:
-        return environment
-    #
-    # Next is saved file in current working directory.
-    if cwd is None:
-        cwd = os.getcwd()
-    local_default = get_saved_default_environment(cwd=cwd)
-    if local_default not in ['', None]:
-        return local_default
-    #
-    # Lowest priority is the directory path basename.
-    return os.path.basename(cwd)
-
-
-def copyanything(src, dst):
-    """Copy anything from src to dst."""
-    try:
-        copytree(src, dst)
-    except FileExistsError as e:  # noqa
-        pass
-    except OSError as err:
-        if err.errno == os.errno.ENOTDIR:
-            copy(src, dst)
-        else:
-            raise
-    finally:
-        remove_other_perms(dst)
-
-
-def copydescriptions(src, dst):
-    """
-    Just copy the descriptions portion of an environment
-    directory from src to dst.
-    """
-
-    if not dst.endswith('.d'):
-        raise RuntimeError(
-            f"[-] destination '{dst}' is not a descriptions "
-            "('.d') directory")
-    # Ensure destination directory exists.
-    os.makedirs(dst, exist_ok=True)
-    errors = []
-    try:
-        if src.endswith('.d') and os.path.isdir(src):
-            copytree(src, dst)
-        else:
-            raise RuntimeError(
-                f"[-] source '{src}' is not a descriptions "
-                "('.d') directory")
-    except OSError as err:
-        errors.append((src, dst, str(err)))
-    # catch the Error from the recursive copytree so that we can
-    # continue with other files
-    except Error as err:
-        errors.extend(err.args[0])
-    if errors:
-        raise Error(errors)
-    remove_other_perms(dst)
-
-
-def is_valid_environment(env_path, verbose_level=1):
-    """
-    Check to see if this looks like a valid environment directory.
-
-    Args:
-      env_path: Path to candidate directory to test.
-      verbose_level: Verbosity level (pass from app args)
-
-    Returns:
-      A boolean indicating whether the directory appears to be a valid
-      environment directory or not based on contents including a
-      'secrets.json' file or a 'secrets.d' directory.
-    """
-    environment = os.path.split(env_path)[1]
-    contains_expected = False
-    yaml_files = []
-    for root, directories, filenames in os.walk(env_path):
-        if 'secrets.json' in filenames or 'secrets.d' in directories:
-            contains_expected = True
-        if 'secrets.yml' in filenames:
-            yaml_files.append(os.path.join(root, 'secrets.yml'))
-        if root.endswith('secrets.d'):
-            yaml_files.extend([
-                os.path.join(root, filename)
-                for filename in filenames
-                if filename.endswith('.yml')
-            ])
-    for filename in yaml_files:
-        if verbose_level > 1:
-            logger.warning("[!] found '%s'", filename)
-    is_valid = (
-        os.path.exists(env_path)
-        and contains_expected
-        and len(yaml_files) == 0
-    )
-    if len(yaml_files) > 0 and verbose_level > 0:
-        logger.warning(
-            "[!] environment '%s' needs conversion (see 'psec utils yaml-to-json --help')",  # noqa
-            environment)
-    if not is_valid and verbose_level > 1:
-        logger.warning(
-            "[!] environment directory '%s' exists but looks incomplete",
-            env_path)
-    return is_valid
-
-
 def is_generable(secret_type=None):
     """Return boolean for generability of this secret type."""
     generability = {i['Type']: i['Generable'] for i in SECRET_TYPES}
@@ -343,16 +166,18 @@ def generate_secret(secret_type=None, *arguments, **kwargs):
     if secret_type == "boolean":  # nosec
         return None
     if secret_type == 'password':  # nosec
-        return generate_password(unique,
-                                 acrostic,
-                                 numwords,
-                                 case,
-                                 delimiter,
-                                 min_words_length,
-                                 max_words_length,
-                                 min_acrostic_length,
-                                 max_acrostic_length,
-                                 wordfile)
+        return generate_password(
+            unique,
+            acrostic,
+            numwords,
+            case,
+            delimiter,
+            min_words_length,
+            max_words_length,
+            min_acrostic_length,
+            max_acrostic_length,
+            wordfile,
+        )
     if secret_type == 'crypt_6':  # nosec
         return generate_crypt6(unique)
     if secret_type == 'token_hex':  # nosec
@@ -370,37 +195,19 @@ def generate_secret(secret_type=None, *arguments, **kwargs):
     raise TypeError(f"Secret type '{secret_type}' is not supported")
 
 
-class Memoize:
-    """Memoize(fn) - an instance which acts like fn but memoizes its arguments.
-
-       Will only work on functions with non-mutable arguments. Hacked to assume
-       that argument to function is whether to cache or not, allowing all
-       secrets of a given type to be set to the same value.
-    """
-
-    def __init__(self, fn):
-        self.fn = fn
-        self.memo = {}
-
-    def __call__(self, *args):
-        if args[0] is True:
-            return self.fn(*args)
-        if args not in self.memo:
-            self.memo[args] = self.fn(*args)
-        return self.memo[args]
-
-
 @Memoize
-def generate_password(unique,
-                      acrostic,
-                      numwords,
-                      case,
-                      delimiter,
-                      min_words_length,
-                      max_words_length,
-                      min_acrostic_length,
-                      max_acrostic_length,
-                      wordfile):
+def generate_password(
+    unique,
+    acrostic,
+    numwords,
+    case,
+    delimiter,
+    min_words_length,
+    max_words_length,
+    min_acrostic_length,
+    max_acrostic_length,
+    wordfile,
+):
     """Generate an XKCD style password"""
 
     # Create a wordlist from the default wordfile.
@@ -433,7 +240,7 @@ def generate_password(unique,
 def generate_crypt6(unique=False, password=None, salt=None):
     """Generate a crypt() style SHA512 ("$6$") digest"""
     try:  # Python 3
-        import crypt
+        import crypt  # pylint: disable=import-outside-toplevel
     except ModuleNotFoundError as e:  # noqa
         raise
     if password is None:
@@ -510,9 +317,9 @@ class SecretsEnvironment(object):
     """
     Class for handling secrets environment metadata.
 
-    Provides an interface to the directory contents for a secrets
-    environment, including groups descriptions, a tmp/ directory,
-    and any other required directories.
+    Provides an interface to the directory contents for a secrets environment,
+    including groups descriptions, a tmp/ directory, and any other required
+    directories.
 
       Typical usage example::
 
@@ -536,39 +343,75 @@ class SecretsEnvironment(object):
 
     logger = logging.getLogger(__name__)
 
-    def __init__(self,
-                 environment=None,
-                 secrets_basedir=None,
-                 secrets_file=os.getenv('D2_SECRETS_BASENAME',
-                                        'secrets.json'),
-                 create_root=True,
-                 defer_loading=True,
-                 export_env_vars=False,
-                 preserve_existing=False,
-                 env_var_prefix=None,
-                 source=None,
-                 verbose_level=1):
-        self._changed = False
-        if environment is not None:
-            self._environment = environment
-        else:
-            self._environment = get_default_environment()
-        self._secrets_file = secrets_file
-        self._secrets_basedir = secrets_basedir
-        self._verbose_level = verbose_level
-        # Ensure root directory exists in which to create secrets
-        # environments?
-        if not self.secrets_basedir_exists():
+    def __init__(
+        self,
+        environment=None,
+        secrets_basedir=None,
+        secrets_file=None,
+        create_root=False,
+        defer_loading=False,
+        export_env_vars=False,
+        preserve_existing=False,
+        env_var_prefix=None,
+        source=None,
+        verbose_level=1,
+    ):
+        """
+        Initialize secrets environment object.
+        """
+        if secrets_file and secrets_basedir:
+            raise RuntimeError(
+                "[-] 'secrets_file' and 'secrets_basedir' are mutually "
+                "exclusive when initializing a SecretsEnvironment()"
+            )
+        self._environment = (
+            get_default_environment() if environment is None
+            else environment
+        )
+        if secrets_basedir is None:
+            secrets_basedir = get_default_secrets_basedir()
+        try:
+            is_secrets_basedir(basedir=secrets_basedir, raise_exception=True)
+            self._secrets_basedir = Path(secrets_basedir)
+        except BasedirNotFoundError:
             if create_root:
-                self.secrets_basedir_create()
+                self._secrets_basedir = secrets_basedir_create(
+                    basedir=secrets_basedir
+                )
             else:
+                raise
+        if secrets_file is not None:
+            self._secrets_file = Path(secrets_file)
+            if len(self._secrets_file.parts) < 3:
+                # This might not exactly be true in all cases, but I don't
+                # have time to run down all possible use cases (or put in
+                # all necessary checks) at the moment.
                 raise RuntimeError(
-                    f"[-] directory '{self.secrets_basedir()}' "
-                    "does not exist and create_root=False")
-        self._secrets_descriptions = f"{os.path.splitext(self._secrets_file)[0]}.d"  # noqa
+                    '[-] the path to a secrets file should have '
+                    f"at least 3 components ('{self._secrets_file}')"
+                )
+            if not self._secrets_file.exists():
+                raise RuntimeError(
+                    f"[-] the specified secrets file '{secrets_file}' "
+                    'does not exist'
+                )
+            if self._environment != self._secrets_file.parts[-2]:
+                raise RuntimeError(
+                    f"[-] environment name '{environment}' does not "
+                    f"match secrets file path '{secrets_file}'"
+                )
+            if Path(secrets_basedir) not in self._secrets_file.parents:
+                raise RuntimeError(
+                    f"[-] base directory '{secrets_basedir}' does not "
+                    f"match secrets file path '{secrets_file}'"
+                )
+            self.secrets_basedir = self._secrets_file.parents[1]
+        else:
+            self._secrets_file = self._secrets_basedir / self._environment / SECRETS_FILE  # noqa
+        self._secrets_descriptions = self._secrets_file.parent / SECRETS_DESCRIPTIONS_DIR # noqa
+        self._verbose_level = verbose_level
         self.export_env_vars = export_env_vars
         self.preserve_existing = preserve_existing
-
         # When exporting environment variables, include one that specifies the
         # environment from which these variables were derived. This also works
         # around a limitation in Ansible where the current working directory
@@ -576,10 +419,10 @@ class SecretsEnvironment(object):
         # script needs to communicate this back to python_secrets in order for
         # it's .python_secrets_environment file to be used to identify the
         # proper environment.)
-
         if self.export_env_vars is True:
+            os.environ['D2_ENVIRONMENT'] = self._environment
+            # Deprecating this variable name:
             os.environ['PYTHON_SECRETS_ENVIRONMENT'] = self._environment
-
         self.env_var_prefix = env_var_prefix
         # Secrets attribute maps; anything else throws exception
         for a in SECRET_ATTRIBUTES:
@@ -589,6 +432,7 @@ class SecretsEnvironment(object):
             self.read_secrets_descriptions()
         self._secrets = OrderedDict()
         self._descriptions = OrderedDict()
+        self._changed = False
 
     def __str__(self):
         """Produce string representation of environment identifier"""
@@ -603,61 +447,20 @@ class SecretsEnvironment(object):
         """Return boolean reflecting changed secrets."""
         return self._changed
 
-    @classmethod
-    def permissions_check(cls, basedir='.', verbose_level=0):
-        """Check for presense of pernicious overly-permissive permissions."""
-        # File permissions on Cygwin/Windows filesystems don't work the
-        # same way as Linux. Don't try to change them.
-        # TODO(dittrich): Is there a Better way to handle perms on Windows?
-        fs_type = get_fs_type(basedir)
-        if fs_type in ['NTFS', 'FAT', 'FAT32']:
-            msg = (f"[-] {basedir} has file system type '{fs_type}': "
-                   "skipping permissions check")
-            cls.logger.info(msg)
-            return False
-        any_other_perms = stat.S_IROTH | stat.S_IWOTH | stat.S_IXOTH
-        for root, dirs, files in os.walk(basedir, topdown=True):
-            for name in files:
-                path = os.path.join(root, name)
-                try:
-                    st = os.stat(path)
-                    perms = st.st_mode & 0o777
-                    open_perms = (perms & any_other_perms) != 0
-                    if (open_perms and verbose_level >= 1):
-                        print(f"[!] file '{path}' is mode {oct(perms)}",
-                              file=sys.stderr)
-                except OSError:
-                    pass
-                for name in dirs:
-                    path = os.path.join(root, name)
-                    try:
-                        st = os.stat(path)
-                        perms = st.st_mode & 0o777
-                        open_perms = (perms & any_other_perms) != 0
-                        if (open_perms and verbose_level >= 1):
-                            print((
-                                    f"[!] directory '{path}' is mode "
-                                    f"{oct(perms)}"
-                                  ),
-                                  file=sys.stderr)
-                    except OSError:
-                        pass
-
     # TODO(dittrich): FIX Cere call
-    def secrets_descriptions_dir(self):
+    def get_secrets_descriptions_dir(self):
         """Return the path to the drop-in secrets description directory"""
         _env = self._environment
         if not _env:
-            return self.secrets_basedir()
+            return self.get_secrets_basedir()
         else:
-            return os.path.join(self.secrets_basedir(),
-                                self.secrets_basename().replace('.json', '.d'))
+            return self.get_secrets_basedir() / self.get_secrets_basename().replace('.json', '.d')  # noqa
 
-    def secrets_basename(self):
+    def get_secrets_basename(self):
         """Return the basename of the current secrets file"""
         return os.path.basename(self._secrets_file)
 
-    def secrets_basedir(self, init=False):
+    def get_secrets_basedir(self, init=False, mode=DEFAULT_MODE):
         """
         Returns the directory path root for secrets storage and definitions.
 
@@ -669,78 +472,72 @@ class SecretsEnvironment(object):
         If the environment variable "D2_SECRETS_BASEDIR" is set, that location
         is used instead.
         """
-        if self._secrets_basedir is None:
-            _home = os.path.expanduser('~')
-            _secrets_subdir = os.path.join(
-                _home, 'secrets' if '\\' in _home else '.secrets')
-            self._secrets_basedir = os.getenv('D2_SECRETS_BASEDIR',
-                                              _secrets_subdir)
-            if not os.path.exists(self._secrets_basedir) and init:
-                self.secrets_basedir_create()
-        return self._secrets_basedir
+        try:
+            secrets_basedir = self._secrets_basedir
+        except AttributeError:
+            secrets_basedir = get_default_secrets_basedir()
+        if init:
+            secrets_basedir_create(
+                basedir=secrets_basedir,
+                mode=mode,
+            )
+        else:
+            is_secrets_basedir(basedir=secrets_basedir, raise_exception=True)
+        return secrets_basedir
 
     def secrets_basedir_exists(self):
         """Return whether secrets root directory exists"""
-        _secrets_basedir = self.secrets_basedir()
-        return os.path.exists(_secrets_basedir)
+        return self._secrets_basedir.exists()
 
-    def secrets_basedir_create(self, mode=DEFAULT_MODE):
-        """Create secrets root directory"""
-        os.makedirs(self.secrets_basedir(),
-                    exist_ok=True,
-                    mode=mode)
-
-    def environment_path(self, env=None, subdir=None, host=None):
+    def get_environment_path(self, env=None, subdir=None, host=None):
         """Returns the absolute path to secrets environment directory
         or subdirectories within it"""
         if env is None:
             env = self._environment
-        _path = os.path.join(self.secrets_basedir(), env)
-
+        _path = self.get_secrets_basedir() / env
         if not (subdir is None and host is None):
             valid_subdir = r'a-zA-Z0-9_/'
             invalid_subdir = re.compile('[^{}]'.format(valid_subdir))
             valid_host = r'a-zA-Z0-9_\./'  # noqa
             invalid_host = re.compile('[^{}]'.format(valid_host))
-
             if subdir is None and host is not None:
                 raise RuntimeError(
                     '[-] Must specify subdir when specifying host')
-
             if subdir is not None:
                 if subdir.startswith('/'):
                     raise RuntimeError('[-] subdir may not start with "/"')
                 elif subdir.endswith('/'):
                     raise RuntimeError('[-] subdir may not end with "/"')
                 if not bool(invalid_subdir.search(subdir)):
-                    _path = os.path.join(_path, subdir)
+                    _path = _path / subdir
                 else:
-                    raise RuntimeError("[-] invalid character in subdir: "
-                                       f"must be in [{valid_subdir}]")
-
+                    raise RuntimeError(
+                        "[-] invalid character in subdir: "
+                        f"must be in [{valid_subdir}]"
+                    )
             if host is not None:
                 if not bool(invalid_host.search(host)):
-                    _path = os.path.join(_path, host)
+                    _path = _path / host
                 else:
                     raise RuntimeError(
                         "[-] invalid character in host: "
                         f"must be in [{valid_host}]")
-
-        return _path
+        return Path(_path)
 
     def environment_exists(self, env=None, path_only=False):
         """Return whether secrets environment directory exists
-        and contains files"""
-        _ep = self.environment_path(env=env)
-        result = os.path.isdir(self.descriptions_path())
+        and contains files other than 'tmp' directory."""
+        _ep = self.get_environment_path(env=env)
+        result = os.path.isdir(self.get_descriptions_path())
         if not result and os.path.exists(_ep):
             if path_only:
                 result = True
             else:
                 _files = list()
-                for root, directories, filenames in os.walk(_ep):
+                for root, _, filenames in os.walk(_ep):
                     for filename in filenames:
-                        _files.append(os.path.join(root, filename))
+                        if filename != 'tmp':
+                            _files.append(os.path.join(root, filename))
                 result = len(_files) > 0
         return result
 
@@ -749,7 +546,7 @@ class SecretsEnvironment(object):
                            alias=False,
                            mode=DEFAULT_MODE):
         """Create secrets environment directory"""
-        env_path = self.environment_path()
+        env_path = self.get_environment_path()
         if not alias:
             # Create a new environment (optionally from an
             # existing environment)
@@ -773,19 +570,17 @@ class SecretsEnvironment(object):
             # Create a symlink with a relative path
             os.symlink(str(source_env), env_path)
 
-    def secrets_file_path(self, env=None):
+    def get_secrets_file_path(self, env=None):
         """Returns the absolute path to secrets file"""
         if env is None:
             env = self._environment
-        return os.path.join(self.secrets_basedir(),
-                            env,
-                            self._secrets_file)
+        return self.get_secrets_basedir() / env / self._secrets_file
 
     def secrets_file_path_exists(self):
         """Return whether secrets file exists"""
-        return os.path.exists(self.secrets_file_path())
+        return os.path.exists(self.get_secrets_file_path())
 
-    def descriptions_path(
+    def get_descriptions_path(
         self,
         root=None,
         group=None,
@@ -794,22 +589,35 @@ class SecretsEnvironment(object):
     ):
         """Return path to secrets descriptions directory or file."""
         if root is not None:
-            path = os.path.join(os.path.abspath(root),
-                                self._secrets_descriptions)
+            path = Path(root) / self._secrets_descriptions
         else:
-            path = os.path.join(self.environment_path(),
-                                self._secrets_descriptions)
+            path = self.get_environment_path() / self._secrets_descriptions
         if create:
-            os.makedirs(path,
-                        exist_ok=True,
-                        mode=mode)
+            path.mkdir(parents=True, exist_ok=True, mode=mode)
         if group is not None:
-            path = os.path.join(path, f"{group}.json")
+            path = path / f"{group}.json"
         return path
 
-    def tmpdir_path(self):
+    def get_tmpdir_path(self, create_path=False):
         """Return the absolute path to secrets descriptions tmp directory"""
-        return os.path.join(self.environment_path(), "tmp")
+        tmpdir = self.get_environment_path() / "tmp"
+        if create_path:
+            tmpdir_mode = 0o700
+            try:
+                os.makedirs(tmpdir, tmpdir_mode)
+                self.logger.info("[+] created tmpdir %s", tmpdir)
+            except FileExistsError:
+                mode = os.stat(tmpdir).st_mode
+                current_mode = S_IMODE(mode)
+                if current_mode != tmpdir_mode:
+                    os.chmod(tmpdir, tmpdir_mode)
+                    self.logger.info(
+                        "[+] changed mode on %s from %s to %s",
+                        oct(current_mode),
+                        oct(tmpdir_mode),
+                        tmpdir
+                    )
+        return tmpdir
 
     def requires_environment(self, path_only=False):
         """
@@ -840,7 +648,7 @@ class SecretsEnvironment(object):
             string: The value of the secret
 
         Raises:
-            RuntimeError: If value is ``None`` and
+            SecretNotFoundError: If value is ``None`` and
                 ``allow_none`` is ``False``
 
         """
@@ -848,7 +656,7 @@ class SecretsEnvironment(object):
             raise RuntimeError('[-] must specify secret to get')
         v = self._secrets.get(secret, None)
         if v is None and not allow_none:
-            raise RuntimeError(f"[-] '{secret}' is not defined")
+            raise SecretNotFoundError(f"[-] '{secret}' is not defined")
         return v
 
     def get_secret_export(self, secret):
@@ -908,7 +716,7 @@ class SecretsEnvironment(object):
         :return:
         """
         try:
-            del(self.Variable[secret])
+            del self.Variable[secret]
         except KeyError:
             pass
         else:
@@ -918,6 +726,18 @@ class SecretsEnvironment(object):
         """Return type for variable or None if no description"""
         return self.Type.get(variable, None)
 
+    def get_default_value(self, variable):
+        """Return the default value from the Options attribute"""
+        try:
+            values = self.Options.get(variable).split(',')
+        except AttributeError:
+            values = []
+        return (
+            values[0]
+            if len(values) > 0 and values[0] != '*'
+            else ''
+        )
+
     def read_secrets_and_descriptions(self):
         """Read secrets descriptions and secrets."""
         self.read_secrets_descriptions()
@@ -956,7 +776,7 @@ class SecretsEnvironment(object):
         descriptions dictionary defined to be None and set self._changed
         to ensure these are written out.
         """
-        _fname = self.secrets_file_path()
+        _fname = self.get_secrets_file_path()
         yaml_fname = f"{os.path.splitext(_fname)[0]}.yml"
         # TODO(dittrich): Add upgrade feature... some day.
         # Until then, reference a way for anyone affected to manually
@@ -986,7 +806,7 @@ class SecretsEnvironment(object):
     def write_secrets(self):
         """Write out the current secrets if any changes were made"""
         if self._changed:
-            _fname = self.secrets_file_path()
+            _fname = self.get_secrets_file_path()
             self.logger.debug("[+] writing secrets to '%s'", _fname)
             with open(_fname, 'w') as f:
                 json.dump(self.Variable, f, indent=2)
@@ -1018,8 +838,8 @@ class SecretsEnvironment(object):
                 f"a '.json' extension ('{src}')")
         if not (os.path.exists(src) or self.environment_exists(env=src)):
             raise RuntimeError(
-                f"[-] '{src}' does not exist")
-        dest = self.descriptions_path(create=True)
+                f"[-] directory or environment '{src}' does not exist")
+        dest = self.get_descriptions_path(create=True)
         if src.endswith('.d'):
             # Copy anything when cloning from directory.
             src_files = get_files_from_path(src)
@@ -1032,7 +852,7 @@ class SecretsEnvironment(object):
             # Only copy descriptions when cloning from environment.
             src_env = SecretsEnvironment(environment=src)
             copydescriptions(
-                src_env.descriptions_path(),
+                src_env.get_descriptions_path(),
                 dest
             )
         remove_other_perms(dest)
@@ -1049,7 +869,7 @@ class SecretsEnvironment(object):
         """
         if group is not None:
             # raise RuntimeError('[!] no group specified')
-            infile = self.descriptions_path(group=group)
+            infile = self.get_descriptions_path(group=group)
         if infile is None:
             raise RuntimeError(
                 '[!] must specify an existing group or file to read')
@@ -1072,10 +892,14 @@ class SecretsEnvironment(object):
         """Write out the secrets descriptions to a file."""
         if group is None:
             raise RuntimeError('[!] no group specified')
-        outfiles = [self.descriptions_path(group=group)]
+        outfiles = [self.get_descriptions_path(group=group)]
         if mirror_to is not None:
-            outfiles.append(self.descriptions_path(root=mirror_to,
-                                                   group=group))
+            outfiles.append(
+                self.get_descriptions_path(
+                    root=mirror_to,
+                    group=group,
+                )
+            )
         for outfile in outfiles:
             os.makedirs(os.path.dirname(outfile),
                         exist_ok=True,
@@ -1084,7 +908,7 @@ class SecretsEnvironment(object):
                 f.write(json.dumps(data, indent=2))
                 f.write('\n')
 
-    def check_duplicates(self, data=list()):
+    def check_duplicates(self, data=None):
         """
         Check to see if any 'Variable' dictionary elements in list match
         any already defined variables. If so, raise RuntimeError().
@@ -1092,15 +916,17 @@ class SecretsEnvironment(object):
         :param data: list of dictionaries containing secret descriptions
         :return: None
         """
-        for d in data:
-            v = d.get('Variable')
-            if v in self._secrets:
-                raise RuntimeError(
-                    f"[-] variable '{v}' duplicates an existing variable")
+        if isinstance(data, list):
+            for d in data:
+                v = d.get('Variable')
+                if v in self._secrets:
+                    raise RuntimeError(
+                        f"[-] variable '{v}' duplicates an existing variable"
+                    )
 
     def read_secrets_descriptions(self):
         """Load the descriptions of groups of secrets from a .d directory"""
-        groups_dir = self.descriptions_path()
+        groups_dir = self.get_descriptions_path()
         if not os.path.exists(groups_dir):
             self.logger.info('[-] secrets descriptions directory not found')
         else:
@@ -1162,6 +988,10 @@ class SecretsEnvironment(object):
         """Get the options for setting the secret"""
         return self.Options.get(secret, '*')
 
+    def get_help(self, secret):
+        """Get the help documentation URL for the secret"""
+        return self.Help.get(secret, '*')
+
     def get_prompt(self, secret):
         """Get the prompt for the secret"""
         return self.Prompt.get(secret, secret)
